#!/usr/bin/python
# -*- coding: utf-8 -*-

import os
import sys
import math
import xbmc,xbmcgui
from threading import Thread
from dudehere.routines import *
from dudehere.routines import plugin
from dudehere.routines.window import Window
from dudehere.routines.constants import WINDOW_ACTIONS
from dudehere.routines.transmogrifier import TransmogrifierAPI

	
WINDOW_PREFIX = 'transmogrifier'
CONTEX_ACTION = None
STATUS = enum(PENDING=1, ACTIVE=2, COMPLETE=3, PAUSED=4, FAILED=-1, SAVING=5)
IDS = enum(STATUS=80010)
status_icons = {1: "pending.png", 2: "active.png", 3: "complete.png", -1: "failed.png", 5: "saving.png"}
SKIN_PATH = plugin.get_setting('skin') if plugin.get_setting('skin') else 'Dark'


def format_size(bytes):
	size = float(bytes) / (1024 * 1024)
	if size > 2000:
		size = size / 1024
		size = "%.2f" % size
		unit = 'GB'
	else :
		size = int(size)
		unit = 'MB'
	size = "%s %s" % (size, unit)
	return size

def reset_db():
	c = plugin.dialog_confirm("Reset Database", "Are you sure?", "You will have to restart kodi afterwards.")
	if c:
		DB_FILE = plugin.get_setting('database_sqlite_file')
		plugin.set_setting("database_sqlite_init", "false")
		vfs.rm(DB_FILE, quiet=True)
		
def select_host():
	from dudehere.routines.plugin import Plugin
	plugin = Plugin()
	hosts_file = vfs.join(DATA_PATH, "hosts.json")
	if vfs.exists(hosts_file):
		history = vfs.read_file(hosts_file, json=True)
		hosts = ["http://%s:%s" % (h['host'], h['port']) for h in history]
		host = plugin.dialog_select('Select a Host', hosts)
		if host is not False:
			host = history[host]
			plugin.set_setting('remote_host', host['host'])
			plugin.set_setting('remote_control_port', host['port'])
			plugin.set_setting('remote_auth_pin', host['pin'])
	else:
		plugin.notify("Error", "Host history is empty")
	
	
	
def view_queue():
	TM = TransmogrifierAPI()
	class Queue(xbmcgui.WindowXMLDialog):
		abort_polling = False
		current_id = -1
		def __init__(self, *args, **kwargs):
			xbmcgui.WindowXML.__init__(self)
		
		def poll_queue(self):
			while True:
				if self.abort_polling: break
				results = TM.get_progress()
				if results:
					if self._count != len(results['queue']):
						self.update(results)
					else:
						for q in results['queue']:
							index = results['queue'].index(q)
							if q[0] == results['progress']['id']:
								self.getControl(80050).setVisible(True)
								self.getControl(80050).setPercent(results['progress']['percent'])
								self.getControl(80010).setLabel('Active...')
								self.getControl(80011).setLabel(q[2])
								self.getControl(80012).setLabel(q[7])
								self.getControl(80014).setLabel(format_size(results['progress']['total_bytes']))
								status = "[COLOR green]%s[/COLOR] of [COLOR green]%s[/COLOR] at [COLOR orange]%s KB/s[/COLOR] %s%s" % (format_size(results['progress']['cached_bytes']), format_size(results['progress']['total_bytes']), results['progress']['speed'], results['progress']['percent'], '%')
								self.getControl(80015).setLabel(status)
							
							elif results['progress']['id'] == 0 and results['progress']['id'] == 0:
								self.getControl(80050).setVisible(False)
								self.getControl(80050).setPercent(0)
								self.getControl(80010).setLabel('Waiting...')
								self.getControl(80011).setLabel('')
								self.getControl(80012).setLabel('')
								self.getControl(80014).setLabel('')
								self.getControl(80015).setLabel('')
							status_icon = status_icons[q[3]]
							
							self.getControl(80000).getListItem(index).setProperty("status", str(q[3]))	
							self.getControl(80000).getListItem(index).setLabel(status_icon)	
					xbmc.sleep(1000)
				else:
					break
				
		def onInit(self):
			self.getControl(80050).setVisible(False)
			connection = "[B][COLOR blue]http://%s:%s[/COLOR][/B]"
			if plugin.get_setting('connect_remote') == 'true':
				connection = connection % (plugin.get_setting('remote_host'), plugin.get_setting('remote_control_port'))
				hosts_file = vfs.join(DATA_PATH, "hosts.json")
				if vfs.exists(hosts_file):
					history = vfs.read_file(hosts_file, json=True)
				else:
					history = []
				old_host = False
				for h in history: 
					old_host = (h['host'] == plugin.get_setting('remote_host') and h['port'] == plugin.get_setting('remote_control_port'))
					if old_host: break
				if not old_host:
					history.append({"host": plugin.get_setting('remote_host'), "port": plugin.get_setting('remote_control_port'), "pin": plugin.get_setting('remote_auth_pin')})	
				vfs.write_file(hosts_file, history, mode='w', json=True)
			else:
				connection = connection % ('localhost', plugin.get_setting('control_port'))
			self.getControl(80013).setLabel(connection)
			
			queue = TM.get_queue()
			self.update(queue)
			monitor = Thread(target=self.poll_queue)
			monitor.start()
		
		def show_context(self, index, status):
			global CONTEX_ACTION
			CONTEX_ACTION = None
			status=int(status)
			items = []
			icon_root = vfs.join(ROOT_PATH, 'resources/skins/%s/media' % SKIN_PATH)
			if status == STATUS.PENDING:
				menu = ["Remove", "Change Priority"]
				height = 150
				posY = 240
			elif status == STATUS.ACTIVE or status==STATUS.SAVING:
				menu = ["Abort"]
				height = 75
				posY = 260
			elif status == STATUS.COMPLETE:
				menu = ["Requeue", "Play From Here", "Remove"]
				height = 210
				posY = 230
			elif status == STATUS.PAUSED:
				menu = ["Requeue", "Remove", "Resume"]
				height = 210
				posY = 230
			else:
				menu = ["Requeue", "Remove"]
				height = 150
				posY = 240
			
			for item in menu:
				icon = item.replace(" ", "_").lower()
				icon = vfs.join(icon_root, icon + '.png')
				liz = xbmcgui.ListItem(item, iconImage=icon)
				items.append(liz)
			
			
			class Context(xbmcgui.WindowXMLDialog):
		
				def __init__(self, *args, **kwargs):
					xbmcgui.WindowXML.__init__(self)
					
				def onInit(self):
					self.getControl(85002).setHeight(height)
					self.getControl(85002).setPosition(440, posY)
					self.getControl(85001).addItems(items)
					pass
				
				def onAction(self, action):
					action = action.getId()
					if action in [WINDOW_ACTIONS.ACTION_PREVIOUS_MENU, WINDOW_ACTIONS.ACTION_NAV_BACK]:
						self.close()
		
				def onClick(self, controlID):
					global CONTEX_ACTION
					obj = self.getFocus()
					CONTEX_ACTION = obj.getSelectedItem().getLabel().lower()
					self.close()
		
				def onFocus(self, controlID):
					pass
				
			ctx = Context("context.xml", ROOT_PATH, SKIN_PATH)
			ctx.doModal()
			del ctx
			return CONTEX_ACTION
		
		def update(self, queue):
			items = []
			for item in queue['queue']:
				status = status_icons[item[3]]
				liz = xbmcgui.ListItem(status, iconImage=item[1]+".png")
				liz.setProperty("priority", str(item[6]))
				liz.setProperty("addon", str(item[7]))
				liz.setProperty("host", '' if item[8] is None else str(item[8]))
				liz.setProperty("file_name", str(item[2]))
				liz.setProperty("status", str(item[3]))
				liz.setProperty("file_id", str(item[5]))
				liz.setProperty("id", str(item[0]))
				path = plugin.get_setting('save_directory')
				if item[1] == 'tvshow':
					path = vfs.join( path + 'TV Shows', vfs.clean_file_name(item[2]))
				elif item[1] == 'torrent':
					path = vfs.join( path + 'Torrents', vfs.clean_file_name(item[2]))
				else:
					path = vfs.join(path + 'Movies', vfs.clean_file_name(item[2]))
				liz.setProperty('path', path)
				items.append(liz)
			self.getControl(80000).reset()
			self.getControl(80000).addItems(items)

			self._count = len(items)

			
		def onAction(self, action):
			action = action.getId()
			if action in [WINDOW_ACTIONS.ACTION_PREVIOUS_MENU, WINDOW_ACTIONS.ACTION_NAV_BACK]:
				self.close()
			
			try:
				if action in [WINDOW_ACTIONS.ACTION_SHOW_INFO, WINDOW_ACTIONS.ACTION_CONTEXT_MENU]:
					controlID = self.getFocus().getId()
					if controlID == 80000:
						index = self.getControl(controlID).getSelectedPosition()
						liz = self.getControl(controlID).getListItem(index)
						status = liz.getProperty("status")
						event = self.show_context(index, status)
						id = int(liz.getProperty("id"))
						file_id = liz.getProperty("file_id")
						if event == 'requeue':
							TM.restart(id)
						elif event == 'abort':
							TM.abort(file_id)
						elif event == 'remove':
							TM.delete(int(id))
							self.getControl(controlID).removeItem(index)
						elif event == 'change priority':
							dialog = xbmcgui.Dialog()
							priority = dialog.numeric(0, 'Enter a new Priority', "10")
							TM.change_priority(id, priority)
							self.current_id = -1
							queue = TM.get_queue()
							set_property('force_refresh', "true")
							self.update(queue)
						elif event == 'play from here':
							path = False
							for extension in ['avi', 'mkv', 'mov', 'mp4', 'flv']:
								test = "%s.%s" % (liz.getProperty("path"), extension)
								if vfs.exists(test):
									path = test
									break
							from dudehere.routines.plugin import Plugin	
							if path:
								self.close()
								Plugin().play_url(path)
							else:
								Plugin().notify("Could not play file", liz.getProperty("file_name"))
			except:
				pass
			
		def onClick(self, controlID):

			if controlID in [82003, 82000]: 
				self.close()
			elif controlID==82001:
				ok = TM.clean_queue()
				if ok:
					xbmc.executebuiltin( "ActivateWindow(busydialog)" )
					xbmc.sleep(1000)
					set_property('force_refresh', "true")
					self.current_id = 0
					queue = TM.get_queue()
					self.update(queue)
					xbmc.executebuiltin( "Dialog.Close(busydialog)" )

	
		def onFocus(self, controlID):
			pass

	q = Queue("queue.xml", ROOT_PATH, SKIN_PATH)
	q.doModal()
	q.abort_polling = True
	del q

def first_run():
	plugin.set_setting('setup_run', 'true')
	plugin.set_setting('version', VERSION)
plugin.first_run = first_run

plugin.register('main', view_queue)
plugin.register('reset_db', reset_db)
plugin.register('select_host', select_host)
if __name__ == '__main__':
	plugin.run()

