#!/usr/bin/python
# -*- coding: utf-8 -*-
import xbmc
import urllib2
import uuid
import time
import math
import hashlib
import urlresolver
import json
import socket
import errno
from datetime import datetime
from threading import Thread
from SocketServer import ThreadingMixIn
from dudehere.routines import *
from dudehere.routines import plugin
from resources.lib.database import *
from dudehere.routines.threadpool import ThreadPool, MyPriorityQueue
from BaseHTTPServer import HTTPServer
from resources.lib.common import *
from dudehere.routines.premiumize import PremiumizeAPI
from resources.lib.server import RequestHandler
from resources.lib.transmogrifier import OutputHandler, Transmogrifier

SETTINGS_CHANGED = True

class Server(HTTPServer):
	def get_request(self):
		self.socket.settimeout(5.0)
		result = None
		while result is None:
			try:
				result = self.socket.accept()
			except socket.timeout:
				pass
		result[0].settimeout(1000)
		return result
	
	def _handle_request_noblock(self):
		"""Handle one request, without blocking.

		I assume that select.select has returned that the socket is
		readable before this function was called, so there should be
		no risk of blocking in get_request().
		"""
		try:
			request, client_address = self.get_request()
		except socket.error, e:
			if isinstance(e.args, tuple):
				print "errno is %d" % e[0]
				if e[0] == errno.EPIPE:
					print "Detected remote disconnect"
				else:
					print e
			else:
				print "socket error ", e
		if self.verify_request(request, client_address):
			try:
				self.process_request(request, client_address)
			except:
				self.handle_error(request, client_address)
				self.shutdown_request(request)

class ThreadedHTTPServer(ThreadingMixIn, Server):
	"""Handle requests in a separate thread."""

class Service(xbmc.Player):
	__enable_transmogrifier = False
	__enable_premiumize = False
	__premiumize_has_login = False
	__dl_largest = False
	
	def __init__(self, *args, **kwargs):
		xbmc.Player.__init__(self, *args, **kwargs)
	
	def onPlayBackStarted(self):
		if get_property('streaming.started'):
			plugin.log("Now I'm playing file id: %s " % get_property('file_id'))
	
	def onPlayBackStopped(self):
		if get_property('streaming.started'):
			plugin.log("Now I'm stopped")
			plugin.log("Abort: %s" % get_property('file_id'))
			set_property("abort_id", get_property('file_id'))
			tail_file = vfs.join(WORK_DIRECTORY, get_property('file_id') + '.tail')
			if vfs.exists(tail_file):
				vfs.rm(tail_file,quiet=True)
			time.sleep(0.25)
			clear_property("abort_id")
			clear_property('playing')
			clear_property('file_id')
			clear_property('streaming.started')
			clear_property("streaming.total_bytes")
			clear_property("streaming.total_blocks")
			clear_property("streaming.file_id")
			clear_property("streaming.abort")
			clear_property("streaming.seek_block")
			clear_property('streaming.tail_requested')
	
	def onPlayBackEnded(self):
		self.onPlayBackStopped()
	
	def update_settings(self):
		global SETTINGS_CHANGED
		if SETTINGS_CHANGED:
			self.__enable_transmogrifier = plugin.get_setting('enable_transmogrifier')=='true'
			self.__enable_premiumize = plugin.get_setting('enable_premiumize') != "true"
			self.__premiumize_has_login = plugin.get_setting('PremiumizeMeResolver_login', addon_id='script.module.urlresolver') == 'true'
			self.__dl_largest = plugin.get_setting('torrent_dl_type') != 'Zip'
			SETTINGS_CHANGED = False
			
			

	def poll_queue(self):
		self.DB.connect()
		if self.__enable_transmogrifier is False: return False, False, False, False, False, False, False
		SQL = "SELECT filename, url, id, video_type, raw_url, save_dir FROM queue WHERE status=1 ORDER BY priority DESC, id LIMIT 1"
		row = self.DB.query(SQL)
		if row:
			name = row[0]
			url = row[1]
			video_type = row[3]
			id = row[2]
			raw_url = row[4]
			save_dir = row[5]
			file_id = hashlib.md5(url).hexdigest()
			if raw_url and not url:
				source = urlresolver.HostedMediaFile(url=raw_url)
				url = source.resolve() if source else None
			self.DB.execute("UPDATE queue SET status=2, fileid=? WHERE id=?", [file_id, id])
			self.DB.commit()
			message = "Queued: %s" % name
			plugin.raise_notify(ADDON_NAME, message)
			return name, url, raw_url, id, file_id, video_type, save_dir
		elif self.check_premiumize():
			pm = PremiumizeAPI()
			results = pm.poll()
			next = False
			pending = self.DB.query("SELECT hash from remote_queue WHERE service='premiumize' AND complete=0", force_double_array=True)
			for result in results['transfers']:
				if result['status'] != 'finished': continue
				test = result['hash'].upper()
				for p in pending:
					if p[0] == test:
						next = result
						break
				if next: break

			if next:
				#plugin.log(next)
				hash = next['hash']
				results = pm.browse(hash)
				content = results['content']
				keys = content.keys()
				results = content[keys[0]]
				
				if 'items' in results:
					if self.__dl_largest:
						size = 0
						final = {}
						for child in results['children']:
							data = results['children'][child]
							if int(data['size']) > size:
								size = int(data['size'])
								final = data
						#plugin.log(final)
						filename = final['name']
						url = final['url']
					else:
						#plugin.log(results)
						filename = results['name']
						url = results['zip']
				else:
					#plugin.log(results)
					filename = results['name']
					url = results['url']
					
				self.DB.execute("UPDATE remote_queue SET complete=1 WHERE hash=?", [hash.upper()])
				self.DB.execute("INSERT INTO queue(video_type, source_addon, filename, fileid, url) VALUES('torrent','premiumize.me',?,?,?)", [filename, hash, url])
				self.DB.commit()
			
			return False, False, False, False, False, False, False
		else:
			return False, False, False, False, False, False, False
	
	def check_premiumize(self):
		if self.__enable_premiumize is False:
			return False
		if self.__premiumize_has_login is False: 
			return False
		epoch_time = int(time.time())
		return int(math.fmod(epoch_time, 10)) == 0

	
	def clear_cache(self):
		if plugin.get_setting('clear_cache') == "true":
			plugin.log("Clearing download cache...")
			files = vfs.ls(WORK_DIRECTORY, pattern='state$|temp$|tail$')
			for foo in files[1]:
				path = vfs.join(WORK_DIRECTORY, foo)
				vfs.rm(path, quiet=True)
		plugin.log("Clearing orphaned jobs...")
		self.DB.connect()
		self.DB.execute("UPDATE queue SET status=-1 WHERE status=2")
		self.DB.commit()

		
	def start(self):
		if DB_TYPE == 'sqlite':
			self.DB = MyDatabaseAPI(DB_FILE, connect=False, quiet=True, check_version=False)
		else:
			self.DB = DB
		plugin.log("Service starting...", 1)
		class Monitor(xbmc.Monitor):
			def onSettingsChanged(self):
				global SETTINGS_CHANGED
				SETTINGS_CHANGED = True
			
			
		monitor = Monitor()
		self.clear_cache()
		plugin.log("Waiting to Transmogrify...",1)
		if plugin.get_setting('network_bind') == 'Localhost':
			address = "127.0.0.1"
		else:
			address = "0.0.0.0"
		plugin.log("Launching WebInterface on: %s:%s" % (address, CONTROL_PORT))
		
		server_class = ThreadedHTTPServer
		httpd = server_class((address, CONTROL_PORT), RequestHandler)
		webserver = Thread(target=httpd.serve_forever)
		webserver.start()
		self.DB.connect()
		while not monitor.abortRequested():
			if monitor.waitForAbort(1) or get_property('service.abort'):
				break
			self.update_settings()
			filename, url, raw_url, id, file_id, video_type, save_dir = self.poll_queue()
			if id:
				plugin.log("Starting to Transmogrify: %s" % filename,1)
				self.id=id
				started = time.time()
				set_property("caching.file_id", file_id)
				TM = Transmogrifier(id, url, raw_url, filename, file_id, video_type=video_type, save_dir=save_dir)
				TM.start()
				
				if get_property("abort_all")=="true":
					self.DB.execute("UPDATE queue SET status=-1 WHERE id=?", [self.id])
				else:
					set_property("caching.complete", file_id)
					self.DB.execute("UPDATE queue SET status=3 WHERE id=?", [self.id])
					SQL = 'INSERT OR REPLACE INTO files(hash, path) VALUES(?,?)' if self.DB.db_type =='sqlite' else 'REPLACE  INTO files(hash, path) VALUES(?,?)'
					self.DB.execute(SQL, [hashlib.md5(video_type+filename).hexdigest(), TM.output_file])
					try:
						if NOTIFICATION:
							from subprocess import call
							wav = vfs.join(ROOT_PATH, 'resources/notifications/'+NOTIFICATION+'.wav')
							if PLATFORM.startswith('linux'):
								call(["aplay", wav])
							elif PLATFORM.startswith('darwin'):
								call(["afplay"], wav)
							elif PLATFORM.startswith('win'):
								import winsound
								winsound.PlaySound(wav, winsound.SND_FILENAME)
					except: pass
				self.DB.commit()
				del TM
		
		httpd.socket.close()
		plugin.log("Service stopping...", 1)
		
if __name__ == '__main__':
	TS = Service()
	TS.start()
	TS.DB.disconnect()